@@ -70,7 +70,7 @@ typedef struct {
   gboolean expect_server_error;
   GError *server_error;
   gboolean server_should_close;
-  gboolean server_closed;
+  gboolean server_running;
 
   char buf[128];
   gssize nread, nwrote;
@@ -79,24 +79,26 @@ typedef struct {
 static void
 setup_connection (TestConnection *test, gconstpointer data)
 {
-  GInetAddress *inet;
-  guint16 port;
-
   test->context = g_main_context_default ();
   test->loop = g_main_loop_new (test->context, FALSE);
-
   test->auth_mode = G_TLS_AUTHENTICATION_NONE;
-
-  /* This is where the server listens and the client connects */
-  port = g_random_int_range (50000, 65000);
-  inet = g_inet_address_new_from_string ("127.0.0.1");
-  test->address = G_SOCKET_ADDRESS (g_inet_socket_address_new (inet, port));
-  g_object_unref (inet);
-
-  /* The identity matches the server certificate */
-  test->identity = g_network_address_new ("server.example.com", port);
 }
 
+/* Waits about 10 seconds for @var to be NULL/FALSE */
+#define WAIT_UNTIL_UNSET(var)				\
+  if (var)						\
+    {							\
+      int i;						\
+							\
+      for (i = 0; i < 13 && (var); i++)			\
+	{						\
+	  g_usleep (1000 * (1 << i));			\
+	  g_main_context_iteration (NULL, FALSE);	\
+	}						\
+							\
+      g_assert (!(var));				\
+    }
+
 static void
 teardown_connection (TestConnection *test, gconstpointer data)
 {
@@ -108,50 +110,67 @@ teardown_connection (TestConnection *test, gconstpointer data)
        */
       g_object_add_weak_pointer (G_OBJECT (test->service), (gpointer *)&test->service);
       g_object_unref (test->service);
-      while (test->service)
-	g_main_context_iteration (NULL, FALSE);
+      WAIT_UNTIL_UNSET (test->service);
     }
 
   if (test->server_connection)
     {
-      while (!test->server_closed)
-	g_main_context_iteration (NULL, FALSE);
+      WAIT_UNTIL_UNSET (test->server_running);
 
-      g_assert (G_IS_TLS_SERVER_CONNECTION (test->server_connection));
       g_object_add_weak_pointer (G_OBJECT (test->server_connection),
 				 (gpointer *)&test->server_connection);
       g_object_unref (test->server_connection);
-      while (test->server_connection)
-	g_main_context_iteration (NULL, FALSE);
+      WAIT_UNTIL_UNSET (test->server_connection);
     }
 
   if (test->client_connection)
     {
-      g_assert (G_IS_TLS_CLIENT_CONNECTION (test->client_connection));
       g_object_add_weak_pointer (G_OBJECT (test->client_connection),
 				 (gpointer *)&test->client_connection);
       g_object_unref (test->client_connection);
-      while (test->client_connection)
-	g_main_context_iteration (NULL, FALSE);
+      WAIT_UNTIL_UNSET (test->client_connection);
     }
 
   if (test->database)
     {
-      g_assert (G_IS_TLS_DATABASE (test->database));
       g_object_add_weak_pointer (G_OBJECT (test->database),
 				 (gpointer *)&test->database);
       g_object_unref (test->database);
-      while (test->database)
-	g_main_context_iteration (NULL, FALSE);
+      WAIT_UNTIL_UNSET (test->database);
     }
 
-  g_object_unref (test->address);
-  g_object_unref (test->identity);
+  g_clear_object (&test->address);
+  g_clear_object (&test->identity);
   g_main_loop_unref (test->loop);
   g_clear_error (&test->read_error);
   g_clear_error (&test->server_error);
 }
 
+static void
+start_server (TestConnection *test)
+{
+  GInetAddress *inet;
+  GSocketAddress *addr;
+  GInetSocketAddress *iaddr;
+  GError *error = NULL;
+
+  inet = g_inet_address_new_from_string ("127.0.0.1");
+  addr = g_inet_socket_address_new (inet, 0);
+  g_object_unref (inet);
+
+  g_socket_listener_add_address (G_SOCKET_LISTENER (test->service), addr,
+                                 G_SOCKET_TYPE_STREAM, G_SOCKET_PROTOCOL_TCP,
+                                 NULL, &test->address, &error);
+  g_assert_no_error (error);
+
+  /* The hostname in test->identity matches the server certificate. */
+  iaddr = G_INET_SOCKET_ADDRESS (test->address);
+  test->identity = g_network_address_new ("server.example.com",
+					  g_inet_socket_address_get_port (iaddr));
+
+  test->server_running = TRUE;
+}
+
 static gboolean
 on_accept_certificate (GTlsClientConnection *conn, GTlsCertificate *cert,
                        GTlsCertificateFlags errors, gpointer user_data)
@@ -196,7 +215,7 @@ on_server_close_finish (GObject        *object,
     g_assert (error != NULL);
   else
     g_assert_no_error (error);
-  test->server_closed = TRUE;
+  test->server_running = FALSE;
 }
 
 static void
@@ -268,14 +287,8 @@ static void
 start_async_server_service (TestConnection *test, GTlsAuthenticationMode auth_mode,
                             gboolean should_close)
 {
-  GError *error = NULL;
-
   test->service = g_socket_service_new ();
-  g_socket_listener_add_address (G_SOCKET_LISTENER (test->service),
-                                 G_SOCKET_ADDRESS (test->address),
-                                 G_SOCKET_TYPE_STREAM, G_SOCKET_PROTOCOL_TCP,
-                                 NULL, NULL, &error);
-  g_assert_no_error (error);
+  start_server (test);
 
   test->auth_mode = auth_mode;
   g_signal_connect (test->service, "incoming", G_CALLBACK (on_incoming_connection), test);
@@ -358,20 +371,14 @@ run_echo_server (GThreadedSocketService *service,
 
   g_io_stream_close (test->server_connection, NULL, &error);
   g_assert_no_error (error);
-  test->server_closed = TRUE;
+  test->server_running = FALSE;
 }
 
 static void
 start_echo_server_service (TestConnection *test)
 {
-  GError *error = NULL;
-
   test->service = g_threaded_socket_service_new (5);
-  g_socket_listener_add_address (G_SOCKET_LISTENER (test->service),
-                                 G_SOCKET_ADDRESS (test->address),
-                                 G_SOCKET_TYPE_STREAM, G_SOCKET_PROTOCOL_TCP,
-                                 NULL, NULL, &error);
-  g_assert_no_error (error);
+  start_server (test);
 
   g_signal_connect (test->service, "run", G_CALLBACK (run_echo_server), test);
 }
